//Title: BeTVOut Class
//Platform: BeOS 5
//Version: 2002.02.19
//Description:
//	A class for controlling BeTVOut-supported hardware.
//

//  Copyright (C) 2002 Kevin H. Patterson

#include "BeTVOut.h"

#include "GraphicsDevice.h"


BeTVOut::BeTVOut(GraphicsDevice * const uGraphicsDevice)
:
	iGraphicsDevice(uGraphicsDevice),
	
	iInitCheck(false),
	
	iTVOut(false),

	hTVOut(-1),
	iTVOutChipData(new riva_data_t),
	
	iTVOutChipType(TVOutChip_Unknown),
	iBTType(0xFF),
	iBTVersion(0xFF),
	iCHBus(0),
	iCHAddr(0),
	iCHVersion(0),
	iOutputDetection(0),
	
	iFormat(Fmt_NTSC),
	iResolution(Res_640_480),
	iOutputAutoDetect(true),
	iOutput(Out_ALL),
	iFiltering(true),
	iColorBars(false),
	iOverscan(false),
	iDualDisplay(false)
{

	int bus0wired = 0, bus1wired = 0,				//no wired busses found
		bus0sane = 0, bus1sane = 0,						//no sane buses
		stat = B_OK;

	hTVOut = open( "/dev/video/riva_iic/1", O_RDWR );
	//check if driver could be reached...
	if (hTVOut < 0)
		return;

	//setup 'magic value' to unlock driver: 
	//(to keep compatibility & prevent crashes with older versions of driver and app...)
	iTVOutChipData->drv_version = 0xFF;	
	//notify driver of our iBTVersion:
	//(to prevent old apps from accessing new drivers...)
	iTVOutChipData->app_version = static_cast<uint16>(iBTVersion);
	//check iBTVersion of driver...
	stat = ioctl( hTVOut, rivaVersion, iTVOutChipData );
	if (iTVOutChipData->drv_version == 0xFF) {
		Log( "riva_iic driver V0.00\n" );
		Log( "\nIncompatible driver iBTVersion (V0.22 or higher needed), exiting...\n" );
		close( hTVOut );
		return;
	}
	//this works without driver supporting riva_data_t struct:
	//well, it did before I changed the first var-type.. :-(
	Log( "riva_iic driver V%.2f\n", iTVOutChipData->drv_version/100.0 ); 
	if (iTVOutChipData->drv_version < 22) {
		Log( "\nIncompatible driver iBTVersion (V0.22 or higher needed), exiting...\n" );
		close( hTVOut );
		return;
	}
	//(now we know the driver supports the riva_data_t struct...)
	if (stat) { //if ioctl returned 'NACK'..
		Log( "\nBeTVOut application too old for use with this driver, exiting...\n"
				"(iBTVersion %.3f or higher needed.)\n", iTVOutChipData->app_version/1000.0 );
		close( hTVOut );
		return;
	}

	//unlock RIVA extended VGA registers...
	ioctl(hTVOut, rivaUnlock);

	//check for one or two available (wired) IIC busses.
	//On RIVA TNT and later, two could be there...
	int RIVAexists(0);
	Log( "RIVA IICbus 0: " );
	iTVOutChipData->iic.busnr = 0;
	if (!ioctl( hTVOut, rivaChkBus, iTVOutChipData )) {
		Log( "detected.\n" );
		RIVAexists = 1;
		bus0wired = 1;
	}	else {
		Log( "not detected.\n" );
	}
	Log( "RIVA IICbus 1: " );
	iTVOutChipData->iic.busnr = 1;
	if (!ioctl( hTVOut, rivaChkBus, iTVOutChipData )) {
		Log( "detected.\n" );
		RIVAexists = 1;
		bus1wired = 1;
	} else {
		Log( "not detected.\n" );
	}
	if (!RIVAexists) {
		Log( "No RIVA IIC bus found, exiting...\n" );
		close( hTVOut );
		return;
	}

	//find at least one sane IIC bus... (test this by sending an unused adress...)
	if (!FindSaneBus(
			hTVOut, bus0wired, bus1wired, &bus0sane, &bus1sane, iTVOutChipData )) {
		Log( "Unexpected results, exiting...\n" );
		close( hTVOut );
		return;
	}

	//now find a TV-out chip...

	//check for BT or CX chip:
	if (FindBT( hTVOut, iTVOutChipData ))
		iTVOutChipType = TVOutChip_BT;
		
	if (iTVOutChipType == TVOutChip_BT)
	{
		//read chiptype and iBTVersion number:
		switch (ReadType( hTVOut, &iBTVersion, &iBTType, iTVOutChipData )) {
		case 0:
			//no errors found:
			Log( "ChipID = %d, chip iBTVersion = %d: Chiptype is ", iBTType, iBTVersion );
			switch (iBTType) {
			case 0:
				Log( "bt868.\n" );
				break;
			case 1:
				Log( "bt869.\n" );
				break;
			case 2:
				Log( "cx25870.\n" );
				break;
			case 3:
				Log( "cx25871.\n" );
				break;
			default:
				Log( "unknown, exiting...\n" );
				close( hTVOut );
				return;
				break;
			}
			break;
		case 1:
			//IIC error encountered:
			close( hTVOut );
			return;
			break;
		default:
			//The chip apparantly is not a TVout chip:
			iTVOutChipType = TVOutChip_Unknown;
			break;
		} 
	}		

	//check for CH chip:
	if (iTVOutChipType == TVOutChip_Unknown)
	{
		if (bus0sane) {
			iTVOutChipData->iic.busnr = 0;
			iCHBus = 0;
			if (ch_detect( hTVOut, iCHBus, &iCHAddr, &iCHVersion, iTVOutChipData ))
				iTVOutChipType = TVOutChip_CH;
		}
		if (bus1sane && (iTVOutChipType == TVOutChip_Unknown)) {
			iTVOutChipData->iic.busnr = 1;
			iCHBus = 1;
			if (ch_detect( hTVOut, iCHBus, &iCHAddr, &iCHVersion, iTVOutChipData ))
				iTVOutChipType = TVOutChip_CH;
		}
		if (iTVOutChipType == TVOutChip_CH) {
			Log( "\nFound one on bus %d, adress $%x;\n", iCHBus, iCHAddr );
			Log( "Chiptype is ch%s.\n", ch_model_name(iCHVersion) );
		}
		else {	//no TVout chip at all...
			Log( "\nNo supported TVout chip found, exiting...\n" );
			close( hTVOut );
			return;
		}
	}

	iTVOutChipData->devID = 0;
	//get current RIVA display mode:
	ioctl( hTVOut, rivaGetMode, iTVOutChipData );

	Log( "(RIVA h_display: %d, ", iTVOutChipData->displ_mode.displ_timing.h_display );
	Log( "v_display: %d)\n", iTVOutChipData->displ_mode.displ_timing.v_display );

	//RIVA programming should be universal for every chipversion, but
	//the 'horizontal sync phase' setting in the TV-out chip SHOULD be done 
	//seperately for different RIVA chips and videomodes though...
	//(for both CH and BT chips).
	//WARNING: This applies for both VESA and 'driver' (BeScreen) mode!
	Log( "Vendor ID = $%x, Device ID = $%x;\n",
		iTVOutChipData->dev.vendor_id, iTVOutChipData->dev.device_id );
	Log( "Nvidia Corporation" );
	switch (iTVOutChipData->dev.device_id)
	{
		case 0x0020:
			Log( " RIVA TNT1 detected.\n" );
			iTVOutChipData->type = 1;
			break;
		case 0x0028:
			Log( " RIVA TNT2 detected.\n" );
			iTVOutChipData->type = 2;
			break;
		case 0x002D:
			Log( " RIVA TNT2-M64 detected.\n" );
			iTVOutChipData->type = 3;
			break;
		case 0x0110:
			Log( " RIVA GeForce2 MX(200/400) detected.\n" );
			iTVOutChipData->type = 1;
			break;
		default:
			Log( ": Untested card detected!\n" );
	}

	iInitCheck = true;
}


BeTVOut::~BeTVOut() {
	if (hTVOut >= 0)
		close( hTVOut );
	hTVOut = -1;
	delete iTVOutChipData;
	iTVOutChipData = NULL;
}


bool BeTVOut::InitCheck() const {
	return iInitCheck;
}


status_t BeTVOut::TVOut( const bool uEnable ) {

	iTVOut = false;

	unsigned char tMode(0);
	const unsigned char tFiltering((iFiltering > 0) & 0x01);
	
	if (iColorBars) {
		switch (iFormat) {
		case Fmt_NTSC:
			tMode = 1;
			break;
		case Fmt_PAL:
			tMode = 4;
			break;
		}
	} else {
		switch (iFormat) {
		case Fmt_NTSC:
			if (iOverscan) {
				switch (iResolution) {
				case Res_640_480:
					tMode = 11;
					break;
				case Res_720_480:
					tMode = 9;
					break;
				}
			} else {
				switch (iResolution) {
				case Res_640_480:
					tMode = 2;
					break;
				case Res_800_600:
					tMode = 3;
					break;
				}
			}
			break;
		case Fmt_PAL:
			if (iOverscan) {
				switch (iResolution) {
				case Res_720_576:
					tMode = 10;
					break;
				case Res_768_576:
					tMode = 12;
					break;
				}
			} else {
				switch (iResolution) {
				case Res_640_480:
					tMode = 5;
					break;
				case Res_800_600:
					tMode = 6;
					break;
				}
			}
			break;
		}
	}

	if (tMode == 0) {
		Log("ERROR: Invalid TV-Out Mode Selected!");
		return B_ERROR;
	}

	display_mode org_mode, new_mode;
	iGraphicsDevice->GetDisplayMode( &org_mode );	//'save' current mode...
	new_mode = org_mode;

	//Reset the TV-Out chip...
	switch (iTVOutChipType) {
	case TVOutChip_BT:
		//fixme: reset BT should be here...
		//(but: beware of the 'locked SDA' syndrome then!)
		if (KillClock_Blackout_BT( hTVOut, iBTType, iTVOutChipData ))
			return B_ERROR;
		break;
	case TVOutChip_CH:
		//reset CH or use another way to make sure all registers get set the same
		//way non-dependent of how they were set before BeTVOut was started...
		break;
	}

	//Perform Auto Detection on video Outputs...
	pOutputAutoDetect();

	//Program the TV-Out chip...
	if (iColorBars) {

		//This is all the programming that is necessary for color bars!
		switch (iTVOutChipType) {
		case TVOutChip_BT:
			switch (tMode) {
			case 1:
				if (InitBT_NTSC640( hTVOut, iTVOutChipData ))
					return B_ERROR;
				break;
			case 4:
				if (InitBT_PAL800( hTVOut, iTVOutChipData ))
					return B_ERROR;
				break;
			default:
				return B_ERROR;
				break;
			}
			break;
		case TVOutChip_CH:
			return B_ERROR;
			break;
		default:
			return B_ERROR;
			break;
		}
		
	} else {

		//A 'REAL' TV-Out Mode has been selected, so do the programming...
		
		//First, 'slowdown RIVA clock'...
		//makes sure TVout chip is not being overclocked! (VESA640/800 is safe..)
		//NOTE: it looks like this is unnesessary after all..
		switch (iResolution)
		{
			case Res_640_480:
				nonVESAsetVESA640(new_mode);
				break;
			case Res_720_480:
			case Res_720_576:
			case Res_768_576:
			case Res_800_600:
				nonVESAsetVESA800(new_mode);
				break;
			default:
				nonVESAsetVESA640(new_mode);
				break;
		}
		
		switch (iTVOutChipType) {
		case TVOutChip_BT:
			
			{
				//Manual config is non-dependent of the state of the PAL hardware input pin;
				//Also SDA lockups occur when setting EN_XCLK after autoconfig!
				//Make sure PAL_MD=0 for NTSC and PAL_MD = 1 for PAL...
				switch (tMode) {
				case 2:
					if (InitBT_NTSC640( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 3:
					if (InitBT_NTSC800( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 5:
					if (InitBT_PAL640( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 6:
					if (InitBT_PAL800( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 9:
					if (InitBT_NTSC720( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 10:
					if (InitBT_PAL720( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 11:
					if (InitBT_NTSC640_OS( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				case 12:
					if (InitBT_PAL800_OS( hTVOut, iTVOutChipData ))
						return B_ERROR;
					break;
				}
				//modify BT Hphase signal to center TV image...
				if (SetBT_Hphase( hTVOut, tMode, iTVOutChipData ))
					return B_ERROR;
			}
			break;
			
		case TVOutChip_CH:

			{
				//For TNT2 chips horizontal position should be shifted right by 8 pixels.
				//(all values tested on TNT2 and GeForce2 cards.)
				unsigned char hoffset = 2;		//2 x 4 pixels: TNT2 or TNT2 M64 
				if (iTVOutChipData->type == 1) hoffset = 0;	//TNT1 or GeForce2
				
				//Because DVD modes with Chrontel are critical a special correction
				//needs to be done for these modes if RIVA is a TNT1 or GeForce2...
				//(It looks like the non-M64/M64 patch is needed, but this doesn't seem to
				// work here. The non-M64/M64 patch correction is normally visible only on
				// non-overscanning modes, so it only needs to be done there if these
				// modes will be set using BeScreen which currently isn't the case.
				// The currently used direct register programming method *is* using the
				// non-M64/M64 patch.
				// NOTE that the normally used 'hoffset' settings are set for centered TV
				// output, including for the Chrontel DVD modes. Only there we get
				// viewport problems if we don't apply the patch below:)
				if ( iOverscan &
						((iResolution == Res_720_480) | (iResolution == Res_720_576))
						& (iTVOutChipData->type == 1) )
					hoffset = 2;
				
				switch (tMode) {
				case 2:
					//NTSC640 is mode 17, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,17,(0x30 + hoffset),
						0x02,0x01, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 3:
					//NTSC800 resembles mode 24.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,24,(0x3e + hoffset),
						0xff,0x00, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 5:
					//PAL640 resembles mode 19.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,19,(0x38 + hoffset),
						0x00,0x00, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 6:
					//PAL800 is mode 20, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,20,(0x40 + hoffset),
						0x00,0x00, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 9:
					//CH DVD MODES NOTICE:
					//The official CH DVD modes are interlaced input only, while
					//the RIVA cannot output interlaced data. This is why these
					//CH modes (25 and 26) cannot be used..  :-(
					//SO: if you want 'perfect' TV output in non-standard modes:
					//get a card with Brooktree BT868/869 chip mounted on it...
					
					//NTSC720 'resembles' mode 16.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: do not change Hoffset: CH viewport is critical,
					//because we 'display' 720 pixels on a 640 pixel wide viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,16,(0x16 + hoffset),
						0x00,0x01, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 10:
					//PAL720 resembles mode 19.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: be carefull if Hoffset in changed: CH H_overscan is
					//critical, because we display 720 pixels in a 800 pixel
					//viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,19,(0x14 + hoffset),
						0x00,0x00, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 11:
					//NTSC640_OS is mode 16, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,16,(0x22 + hoffset),
						0x00,0x01, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				case 12:
					//PAL768 resembles mode 19.
					ch_rud_set_mode(
						hTVOut, iCHBus, iCHAddr,19,(0x1a + hoffset),
						0x00,0x00, tFiltering, iTVOutChipData);
						//mode, horpos, verpos, oflow
					break;
				}
				//BTW: I checked the videosignals on a scope:
				//all used modes are all perfectly OK!
		
				//BTW 2: The CH700X seem to be much more forgiving than the BT chips!
				//If you change just a little in the wrong way on the BT your TV and VGA
				//image dissappear, while the CH does not suffer from these things!
				//Also the CH pixelclock may be as high as 50Mhz, while BT max is only 40Mhz..
				//The downside: The CH only supports the predefined modes, while the BT
				//is (almost) fully programmable...
		
				//BTW3: Above picture modes and position are tested on a
				//ASUS AGP 32Mb V7100 GeForce2 MX200 with CH7007 TVout chip,
				//and on a TNT2 card.
		
				//modify Hphase signal to center TV image...
				//already done in 'ch_rud_set_mode' above.
		
				//setup CH for TVout (filters...)
			}
			break;
			
		default:
			
			return B_ERROR;
			break;
			
		}

	}

	//Program the RIVA...
	switch (tMode) {
	case 1:
	case 4:
		if (DoTestImage( hTVOut, iTVOutChipData ))
			return B_ERROR;
   	Log(" and NTSC testimage activated.\n");
		break;
	case 2:
		//reprogram RIVA for NTSC 640x480 @ 60Hz mode
		ioctl(hTVOut, rivaNtsc640, iTVOutChipData);
		Log(" and NTSC 640x480 @ 60Hz mode activated.\n");
		break;
	case 3:
		//reprogram RIVA for NTSC 800x600 @ 60Hz mode
		ioctl(hTVOut, rivaNtsc800, iTVOutChipData);
		Log(" and NTSC 800x600 @ 60Hz mode activated.\n");
		break;
	case 5:
		//reprogram RIVA for PAL 640x480 @ 50Hz mode
		ioctl(hTVOut, rivaPal640, iTVOutChipData);
		Log(" and PAL 640x480 @ 50Hz mode activated.\n");
		break;
	case 6:
		//reprogram RIVA for PAL 800x600 @ 50Hz mode
		ioctl(hTVOut, rivaPal800, iTVOutChipData);
		Log(" and PAL 800x600 @ 50Hz mode activated.\n");
		break;
	case 9:
		if (iTVOutChipType == TVOutChip_BT) {
			//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
			//(tested on TNT2 with BT869)
			new_mode.virtual_width = 720;			//H_ACTIVE
			new_mode.timing.h_display = 720;
			new_mode.virtual_height = 480;			//V_ACTIVEI
			new_mode.timing.v_display = 480;

			new_mode.timing.h_sync_start = 744;		//do not change!
			new_mode.timing.h_sync_end = 744+144;	//delta is H_sync_pulse
			new_mode.timing.h_total = 888;			//BT H_TOTAL
			new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
			new_mode.timing.v_sync_end = 490+26;	//delta is V_sync_pulse
			new_mode.timing.v_total = 525;			//CH V_TOTAL
			new_mode.timing.pixel_clock = ((888 * 525 * 60) / 1000); //refresh
		}	else { //CH found
			//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
			//(tested on TNT2-M64 with CH7006)
			new_mode.virtual_width = 720;			//H_ACTIVE
			new_mode.timing.h_display = 720;
			new_mode.virtual_height = 480;			//V_ACTIVEI
			new_mode.timing.v_display = 480;

			new_mode.timing.h_sync_start = 744;		//do not change! (CH viewport is critical...)
			new_mode.timing.h_sync_end = 744+20;	//delta is H_sync_pulse
			new_mode.timing.h_total = 784;			//CH H_TOTAL
			new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
			new_mode.timing.v_sync_end = 490+25;	//delta is V_sync_pulse
			new_mode.timing.v_total = 525;			//CH V_TOTAL
			new_mode.timing.pixel_clock = ((784 * 525 * 60) / 1000); //refresh
		}
		iGraphicsDevice->SetDisplayMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(hTVOut, rivaEnableTVOut);
		Log(" and NTSC 720x480 @ 60Hz overscanning mode activated.\n");
		break;
	case 10:
		//reprogram RIVA for PAL 720x576 @ 50Hz mode:(a bit modified for CH)
		new_mode.virtual_width = 720;			//BT H_ACTIVE
		new_mode.timing.h_display = 720;
		new_mode.virtual_height = 576;			//BT  V_ACTIVEI
		new_mode.timing.v_display = 576;
		new_mode.timing.h_sync_start = 744;		//set for centered sync pulse
		new_mode.timing.h_sync_end = 744+140;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 888;			//BT H_CLKI
		new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 625;			//BT V_LINESI
		new_mode.timing.pixel_clock = ((888 * 625 * 50) / 1000); //refresh
		iGraphicsDevice->SetDisplayMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(hTVOut, rivaEnableTVOut);
		Log(" and PAL 720x576 @ 50Hz overscanning mode activated.\n");
		break;
	case 11:
		//reprogram RIVA for NTSC 640x480 (OverScanning) @ 60Hz mode:
		new_mode.virtual_width = 640;			//BT H_ACTIVE
		new_mode.timing.h_display = 640;
		new_mode.virtual_height = 480;			//BT  V_ACTIVEI
		new_mode.timing.v_display = 480;
		new_mode.timing.h_sync_start = 744;		//set for CH/BT compatible TV output
		new_mode.timing.h_sync_end = 744+20;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 784;			//BT H_CLKI
		new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 490+25;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 525;			//BT V_LINESI
		new_mode.timing.pixel_clock = ((784 * 525 * 60) / 1000); //refresh
		iGraphicsDevice->SetDisplayMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(hTVOut, rivaEnableTVOut);
		Log(" and NTSC 640x480 @ 60Hz overscanning mode activated.\n");
		break;
	case 12:
		//reprogram RIVA for PAL 768x576 @ 50Hz mode:
		new_mode.virtual_width = 768;			//H_ACTIVE
		new_mode.timing.h_display = 768;
		new_mode.virtual_height = 576;			//V_ACTIVEI
		new_mode.timing.v_display = 576;
		new_mode.timing.h_sync_start = 848;		//set for centered TV output
		new_mode.timing.h_sync_end = 848+20;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 944;			//BT H_CLKI
		new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 625;			//BT V_LINESI
		new_mode.timing.pixel_clock = ((944 * 625 * 50) / 1000); //refresh
		iGraphicsDevice->SetDisplayMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(hTVOut, rivaEnableTVOut);
		Log(" and PAL 768x576 @ 50Hz overscanning mode activated.\n");
		break;
	}

	pOutput();
	pDualDisplay();

	iTVOut = uEnable;

	return B_OK;

}


bool BeTVOut::TVOut() const {
	return iTVOut;
}


status_t BeTVOut::Format( const int32 uFormat ) {
	iFormat = uFormat;
	return B_OK;
}


int32 BeTVOut::Format() const {
	return iFormat;
}


status_t BeTVOut::Resolution( const int32 uResolution ) {
	iResolution = uResolution;
	return B_OK;
}


int32 BeTVOut::Resolution() const {
	return iResolution;
}


status_t BeTVOut::OutputAutoDetect( const bool uOutputAutoDetect ) {
	iOutputAutoDetect = uOutputAutoDetect;
	if (iTVOut)
		if (pOutputAutoDetect() >= B_OK) {
			return pOutput();
		}
	return B_ERROR;
}


status_t BeTVOut::pOutputAutoDetect() {
	if (iOutputAutoDetect) {
		switch (iTVOutChipType) {
		case TVOutChip_BT:
			//if TV output signals is set to 'autodetect' then do it:
			iOutputDetection = ReadBT_ConnectStat(hTVOut, iTVOutChipData);
			if (iColorBars) {
				if (DoTestImage( hTVOut, iTVOutChipData ))
					return B_ERROR;
			}
			if (iOutputDetection != -1)
				return B_OK;
			break;
		case TVOutChip_CH:
			//if TV output signals is set to 'autodetect' then do it:
			//(make sure this does only gets called if CH input clock gets set below
			// its speed limits afterwards, or CH might get destroyed...)
			iOutputDetection = ch_detect_connection(hTVOut, iCHBus, iCHAddr, iTVOutChipData);
			Log( "\n(TV output monitor status = %d)\n", iOutputDetection );
			return B_OK;
			//iOutputDetection value is:
			//0: Y/C and CVBS is connected
			//1: only Y/C is connected
			//6: only CVBS is connected
			//7: nothing is connected.
			//(tested with ASUS V7100 GeForce2 MX200 AGP 32Mb.)
			break;
		}
	}
	return B_ERROR;
}


bool BeTVOut::OutputAutoDetect() const {
	return iOutputAutoDetect;
}


status_t BeTVOut::Output( const int32 uOutput ) {
	iOutput = uOutput;
	if (iTVOut)
		return pOutput();
	return B_ERROR;
}


status_t BeTVOut::pOutput() {
	switch (iTVOutChipType) {
	case TVOutChip_BT:
		//setup BT for TVout (filters, DAC_muxes...)
		if (
			!SetupBTforTVout(
				hTVOut, static_cast<unsigned char>(iOutput & 0x03), iOutputDetection,
				static_cast<unsigned char>((iFiltering > 0) & 0x01), iTVOutChipData )
			)
			return B_OK;
		break;
	}
	return B_ERROR;
}


int32 BeTVOut::Output() const {
	return iOutput;
}


status_t BeTVOut::Filtering( const int32 uFiltering ) {
	iFiltering = uFiltering;
	if (iTVOut) {
		return pOutput();
	} else {
		//TVOut is NOT currently enabled
		//The requested state has been stored for the next TVOut mode...
		return B_OK;
	}
	return B_ERROR;
}


int32 BeTVOut::Filtering() const {
	return iFiltering;
}


status_t BeTVOut::ColorBars( const bool uColorBars) {
	iColorBars = uColorBars;
	return B_OK;
}


bool BeTVOut::ColorBars() const {
	return iColorBars;
}


status_t BeTVOut::Overscan( const bool uOverscan ) {
	iOverscan = uOverscan;
	return B_OK;
}


bool BeTVOut::Overscan() const {
	return iOverscan;
}


status_t BeTVOut::DualDisplay( const bool uDualDisplay ) {
	iDualDisplay = uDualDisplay;
	if (iTVOut) {
		return pDualDisplay();
	} else {
		//TVOut is NOT currently enabled
		//The requested state has been stored for the next TVOut mode...
		return B_OK;
	}
	return B_ERROR;
}


status_t BeTVOut::pDualDisplay() {
	if (iDualDisplay) {
		//Turn on VGA output:
		ioctl(hTVOut, rivaVgaOn);
		return B_OK;
	} else {
		//Turn off VGA output:
		ioctl(hTVOut, rivaVgaOff);
		return B_OK;
	}
}


bool BeTVOut::DualDisplay() const {
	return iDualDisplay;
}


int BeTVOut::nonVESAsetVESA640( display_mode new_mode ) {
	//set VESA640: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 640;
	new_mode.virtual_height = 480;
	new_mode.timing.h_display = 640;
	new_mode.timing.v_display = 480;
	
	new_mode.timing.h_sync_start = (0x53 * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1f + (0x53 && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x5F * 8);//byte oriented! (val. *must* be set!)
	new_mode.timing.v_sync_start = 0x1ea;
	new_mode.timing.v_sync_end = ((0x1ea && 0xff0) + 0x00c);
	new_mode.timing.v_total = 0x20b;
	new_mode.timing.pixel_clock = ((((0x5F * 8) * 0x20b)* 60)/1000);// in kHz!

	return iGraphicsDevice->SetDisplayMode( &new_mode );
}


int BeTVOut::nonVESAsetVESA800( display_mode new_mode ) {
	//set VESA800: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 800;
	new_mode.virtual_height = 600;
	new_mode.timing.h_display = 800;
	new_mode.timing.v_display = 600;
	
	new_mode.timing.h_sync_start = (0x6a * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1a + (0x6a && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x7f * 8);//byte oriented! (this value *must* be set!)
	new_mode.timing.v_sync_start = 0x259;
	new_mode.timing.v_sync_end = ((0x259 && 0xff0) + 0x00d);
	new_mode.timing.v_total = 0x272;
	new_mode.timing.pixel_clock = (((((0x7F) * 8) * 0x272)* 60)/1000);// in kHz!

	return iGraphicsDevice->SetDisplayMode( &new_mode );
}
