//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2001.02.03
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson

#include "GraphicsDevice.h"

#include "BeTVOut.h"

#include "Log.h"

#include <iostream.h>

GraphicsDevice::GraphicsDevice(const char * DeviceName)
:
	TVOut(NULL),
	iDeviceName(DeviceName),
	iInitCheck(false),
	hDevice(-1),
	AccelerantImage(-1),
	GetAccelerantFunction(NULL),
	AccelerantDeviceInfo(new accelerant_device_info),
	
	iFramebufferAreaOwner(false)
{
	for (int idx(0); idx < MaxBuffers; ++idx)
		iBufferOffset[idx] = 0;

	//open the driver...
	hDevice = open(DeviceName, B_READ_WRITE);
	if (hDevice >= 0) {
		//device driver has been opened sucessfully...

		//load the accelerant image...
		AccelerantImage = LoadAccelerantImage();
		if (AccelerantImage >= 0) {
			//accelerant image has been loaded successfully...

			iInitCheck = true;

			get_accelerant_device_info GetAccelerantDeviceInfo(NULL);
			GetAccelerantDeviceInfo = (get_accelerant_device_info)GetAccelerantFunction(B_GET_ACCELERANT_DEVICE_INFO, NULL);
			
			if (GetAccelerantDeviceInfo != NULL) {
				//HACK: I have to call this function twice for some STRANGE reason!
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
			}

			iGetMode();
		
		}
		else {
			hDevice = -1;	
		}
	}
	
	TVOut = new BeTVOut( this );
}


GraphicsDevice::~GraphicsDevice() {

	delete TVOut;
	
	delete AccelerantDeviceInfo;

	//unload the accelerant image...
	if (AccelerantImage >= 0) {
		//un-initialize the accelerant...
		uninit_accelerant UnInitAccelerant(NULL);
		UnInitAccelerant = (uninit_accelerant)GetAccelerantFunction(B_UNINIT_ACCELERANT, NULL);
		if (UnInitAccelerant != NULL)
			UnInitAccelerant();
		
		//unload the accelerant image...
		unload_add_on(AccelerantImage);
	}
	
	//close the driver...
	if (hDevice >= 0)
		close(hDevice);
}


bool GraphicsDevice::InitCheck() {
	return iInitCheck;
}


const char * GraphicsDevice::DeviceName() {
	return iDeviceName;
}


const char * GraphicsDevice::Name() {
	return AccelerantDeviceInfo->name;
}


const char * GraphicsDevice::Chipset() {
	return AccelerantDeviceInfo->chipset;
}


const char * GraphicsDevice::SerialNumber() {
	return AccelerantDeviceInfo->serial_no;
}


uint32 GraphicsDevice::Memory() {
	return AccelerantDeviceInfo->memory;
}


uint32 GraphicsDevice::DACSpeed() {
	return AccelerantDeviceInfo->dac_speed;
}


image_id GraphicsDevice::LoadAccelerantImage() {
	status_t result;
	image_id image = -1;
	uint i;
	char
		signature[1024],
		path[PATH_MAX];
	struct stat st;
	static const directory_which vols[] = {
		B_USER_ADDONS_DIRECTORY,
		B_COMMON_ADDONS_DIRECTORY,
		B_BEOS_ADDONS_DIRECTORY
	};

	//get the accelerant signature from the driver...
	result = ioctl(hDevice, B_GET_ACCELERANT_SIGNATURE, &signature, sizeof(signature));
	if (result != B_OK) goto done;

	for (i=0; i < sizeof (vols) / sizeof (vols[0]); i++) {
		//	compute directory path to common or beos addon directory on
		//	floppy or boot volume

  	if (find_directory (vols[i], -1, false, path, PATH_MAX) != B_OK) continue;

		strcat (path, "/accelerants/");
		strcat (path, signature);

		// don't try to load non-existant files
		if (stat(path, &st) != 0) continue;

		// load the image
		image = load_add_on(path);
		if (image >= 0) {

			// get entrypoint from accelerant
			result = get_image_symbol (
				image,
				B_ACCELERANT_ENTRY_POINT,
#if defined(__INTEL__)
				B_SYMBOL_TYPE_ANY,
#else
				B_SYMBOL_TYPE_TEXT,
#endif
				(void **)&GetAccelerantFunction
			);

			if (result == B_OK) {
				init_accelerant ia;
				ia = (init_accelerant)GetAccelerantFunction(B_INIT_ACCELERANT, NULL);
				if (ia && ((result = ia(hDevice)) == B_OK)) {
					// we have a winner!
					break;
				}
			}

			// unload the accelerant, as we must be able to init!
			unload_add_on(image); 
		}

		// mark failure to load image
		image = -1;
	}

done:
	return image;
}


status_t GraphicsDevice::GetMode( display_mode * uDisplayMode ) {
	status_t result(B_ERROR);
	result = iGetMode();
	if (result >= B_OK)
		*uDisplayMode = iDisplayMode;
	return result;
}


status_t GraphicsDevice::iGetMode() {
	get_display_mode Accelerant_GetDisplayMode =
		(get_display_mode)GetAccelerantFunction(B_GET_DISPLAY_MODE, NULL);
	status_t result(B_ERROR);
	result = Accelerant_GetDisplayMode( &iDisplayMode );
	if (result >= B_OK) {
		get_frame_buffer_config gfbc;
		gfbc = (get_frame_buffer_config)GetAccelerantFunction(B_GET_FRAME_BUFFER_CONFIG, NULL);
		if (gfbc != NULL)	gfbc(&iFrameBufferConfig);
	}
	return result;
}


status_t GraphicsDevice::SetMode( display_mode * new_mode, bool = false ) {
	propose_display_mode pdm;
	pdm = (propose_display_mode)GetAccelerantFunction(B_PROPOSE_DISPLAY_MODE, NULL);
	display_mode low, high;
	
	//KP: The following switch was added because I determined, through testing,
	//that propose_mode crashes on certain drivers if called without a valid color space.
	switch (new_mode->space) {
	case B_GRAY8:
		new_mode->space = B_CMAP8;
		break;
	//The following modes have been tested to work on my system...
	case B_CMAP8:
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
	case B_RGB24:
	case B_RGB32:
	case B_RGBA32:
		break;
	default:
		new_mode->space = B_CMAP8;
		break;
	}	
	
	low = *new_mode;
	new_mode->virtual_height = 0xFFFF;  //Make the virtual height as tall as possible
	high = *new_mode;
	if (pdm != NULL) {
		if (pdm( new_mode, &low, &high ) != B_ERROR) {
			set_display_mode sdm;
			sdm = (set_display_mode)GetAccelerantFunction(B_SET_DISPLAY_MODE, NULL);
			if (sdm != NULL) {
				status_t result(B_ERROR);
				result = sdm(new_mode);
				if (result >= B_OK) {
					iDisplayMode = *new_mode;

					get_frame_buffer_config gfbc;
					gfbc = (get_frame_buffer_config)GetAccelerantFunction(B_GET_FRAME_BUFFER_CONFIG, NULL);
					if (gfbc != NULL)	gfbc(&iFrameBufferConfig);
			
					Log("New Mode Set OK!\n");
				}
				return result;
			}
		}
	}
	return B_ERROR;
}


status_t GraphicsDevice::MoveDisplayArea(const int x, const int y) {
	move_display_area mda = (move_display_area)GetAccelerantFunction(B_MOVE_DISPLAY, NULL);
	return mda(x, y);
}


status_t GraphicsDevice::MoveDisplayArea(const int32 uBufferFlags) {
	//cout << "Flag: " << uBufferFlags << endl;
	size_t Offset(0);
	if (uBufferFlags < MaxBuffers)
		Offset = iBufferOffset[uBufferFlags];
	int y = Offset / iFrameBufferConfig.bytes_per_row;
	int x = Offset % iFrameBufferConfig.bytes_per_row;
	//cout << "MDA x y: " << x << " " << y << endl;
	move_display_area mda = (move_display_area)GetAccelerantFunction(B_MOVE_DISPLAY, NULL);
	return mda(x, y);
}


BBufferGroup * GraphicsDevice::CreateBufferGroup(const int uDesiredBufferCount = 8) {

	int tDesiredBufferCount = uDesiredBufferCount;
	if (tDesiredBufferCount > MaxBuffers) tDesiredBufferCount = MaxBuffers;
	
	area_id MyFrameBuffer;
	
	MyFrameBuffer = area_for(iFrameBufferConfig.frame_buffer);
	if (MyFrameBuffer >= 0) {
		Log("Framebuffer Area found.\n");
	} else {
		MyFrameBuffer = create_area("BeTVOut Video framebuffer", &iFrameBufferConfig.frame_buffer, B_EXACT_ADDRESS, 307200 * 4, B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
		if (MyFrameBuffer >=0) {
			iFramebufferAreaOwner = true;
			Log("Framebuffer Area created.\n");
		}
	}
	
	area_info ai;
	if (MyFrameBuffer >=0) {
		Log("Framebuffer Area info:\n");
		get_area_info(MyFrameBuffer, &ai);
		cout << "  Area ID: " << ai.area << endl;
		cout << "  Name: " << ai.name << endl;
		cout << "  Size (K): " << (ai.size / 1024) << endl;
		cout << "  Locked Size (K): " << (ai.ram_size / 1024) << endl;
		cout << "  Team Address: " << ai.address << endl;
		cout << "  FB Address: " << iFrameBufferConfig.frame_buffer << endl;
		cout << "  FB Offset: " << (((char *)iFrameBufferConfig.frame_buffer) - ((char *)ai.address)) << endl;

	} else {
		Log(strerror(MyFrameBuffer));
		Log("ERROR: Framebuffer Area could not be found or created.\n");
		return NULL;
	}
	
	int BufferCount(0);
	BBufferGroup *my_group = new BBufferGroup;

	cout << "Bytes per row: " << iFrameBufferConfig.bytes_per_row << endl;
	cout << "Virtual height: " << iDisplayMode.virtual_height << endl;
	cout << "V display: " << iDisplayMode.timing.v_display << endl;

	size_t BufferOffset = ((char *)iFrameBufferConfig.frame_buffer) - ((char *)ai.address);
	BufferOffset = (BufferOffset + B_PAGE_SIZE - 1) & ~(B_PAGE_SIZE - 1);
	size_t NextOffset = BufferOffset + (iFrameBufferConfig.bytes_per_row * iDisplayMode.timing.v_display);
	NextOffset = (NextOffset + B_PAGE_SIZE - 1) & ~(B_PAGE_SIZE - 1);
	
	while ((NextOffset < ai.size) && (BufferCount < tDesiredBufferCount)) {
		{
			cout << "Buffer Offset: " << BufferOffset << endl;
		
			buffer_clone_info bc_info; 
			bc_info.area = MyFrameBuffer; 
			bc_info.offset = BufferOffset;
			bc_info.size = iFrameBufferConfig.bytes_per_row * iDisplayMode.timing.v_display;

			bc_info.flags = BufferCount;
			iBufferOffset[BufferCount] = BufferOffset;
			
			status_t err = my_group->AddBuffer(bc_info);
			
			if (err >= B_OK)
				BufferCount++;
			else
				break;
		}
		
		BufferOffset = NextOffset;
		NextOffset = BufferOffset + (iFrameBufferConfig.bytes_per_row * iDisplayMode.timing.v_display);
		NextOffset = (NextOffset + B_PAGE_SIZE - 1) & ~(B_PAGE_SIZE - 1);
	}

	return my_group;
}


void GraphicsDevice::DeleteBufferGroup(BBufferGroup * uBufferGroup) {
	delete uBufferGroup;
	uBufferGroup = NULL;
	if (iFramebufferAreaOwner) {
		//Delete the Framebuffer Area, because we created it...
		//BETA: Needs implementation
		//
		//delete_area(MyFrameBuffer);
	}
}
