//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2001.02.03
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson

#include "GraphicsDevice.h"

#include "BeTVOut.h"

#include "Log.h"
	

GraphicsDevice::GraphicsDevice(const char * DeviceName)
:
	TVOut(NULL),
	iDeviceName(DeviceName),
	iInitCheck(false),
	hDevice(-1),
	AccelerantImage(-1),
	GetAccelerantFunction(NULL),
	AccelerantDeviceInfo(new accelerant_device_info)
{
	//open the driver...
	hDevice = open(DeviceName, B_READ_WRITE);
	if (hDevice >= 0) {
		//device driver has been opened sucessfully...

		//load the accelerant image...
		AccelerantImage = LoadAccelerantImage();
		if (AccelerantImage >= 0) {
			//accelerant image has been loaded successfully...

			iInitCheck = true;

			get_accelerant_device_info GetAccelerantDeviceInfo(NULL);
			GetAccelerantDeviceInfo = (get_accelerant_device_info)GetAccelerantFunction(B_GET_ACCELERANT_DEVICE_INFO, NULL);
			
			if (GetAccelerantDeviceInfo != NULL) {
				//HACK: I have to call this function twice for some STRANGE reason!
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
				GetAccelerantDeviceInfo(AccelerantDeviceInfo);
			}

		}
		else {
			hDevice = -1;	
		}
	}
	
	TVOut = new BeTVOut( this );
}

GraphicsDevice::~GraphicsDevice() {

	delete TVOut;
	
	delete AccelerantDeviceInfo;

	//unload the accelerant image...
	if (AccelerantImage >= 0) {
		//un-initialize the accelerant...
		uninit_accelerant UnInitAccelerant(NULL);
		UnInitAccelerant = (uninit_accelerant)GetAccelerantFunction(B_UNINIT_ACCELERANT, NULL);
		if (UnInitAccelerant != NULL)
			UnInitAccelerant();
		
		//unload the accelerant image...
		unload_add_on(AccelerantImage);
	}
	
	//close the driver...
	if (hDevice >= 0)
		close(hDevice);
}

bool GraphicsDevice::InitCheck() {
	return iInitCheck;
}

const char * GraphicsDevice::DeviceName() {
	return iDeviceName;
}

const char * GraphicsDevice::Name() {
	return AccelerantDeviceInfo->name;
}

const char * GraphicsDevice::Chipset() {
	return AccelerantDeviceInfo->chipset;
}

const char * GraphicsDevice::SerialNumber() {
	return AccelerantDeviceInfo->serial_no;
}

uint32 GraphicsDevice::Memory() {
	return AccelerantDeviceInfo->memory;
}

uint32 GraphicsDevice::DACSpeed() {
	return AccelerantDeviceInfo->dac_speed;
}

BBufferGroup * GraphicsDevice::CreateBufferGroup() {

	get_frame_buffer_config gfbc;
	gfbc = (get_frame_buffer_config)GetAccelerantFunction(B_GET_FRAME_BUFFER_CONFIG, NULL);
	frame_buffer_config a_frame_buffer;
	if (gfbc == NULL)
		return NULL;
	if (gfbc(&a_frame_buffer) != B_OK)
		return NULL;

	area_id MyFrameBuffer;
	bool WeCreatedIt(false);

	MyFrameBuffer = area_for(a_frame_buffer.frame_buffer);
	if (MyFrameBuffer < 0) {
		MyFrameBuffer = create_area("MyFrameBuffer", &a_frame_buffer.frame_buffer, B_EXACT_ADDRESS, 307200 * 4, B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
		if (MyFrameBuffer >=0) WeCreatedIt = true;
	}

	if (MyFrameBuffer >=0) {
		Log("We got it.\n");
		area_info ai;
		get_area_info(MyFrameBuffer, &ai);
/*		cout << "Area ID: " << ai.area << endl;
		cout << "Name: " << ai.name << endl;
		cout << "Size (K): " << (ai.size / 1024) << endl;
		cout << "Locked Size (K): " << (ai.ram_size / 1024) << endl;
		cout << "Team Address: " << ai.address << endl; */
	}
	else {
		Log(strerror(MyFrameBuffer));
		Log("\n");
	}

	if (WeCreatedIt) Log("We actually created it.\n");

	if (WeCreatedIt) delete_area(MyFrameBuffer);
	
	//BBufferGroup *my_group = new BBufferGroup; 
/*	BBitmap *my_bitmap = new BBitmap(BRect(0,0,639,479), B_BITMAP_IS_AREA, B_RGB32, 640*4); 
	
	if (!my_bitmap->IsValid()) { 
		return B_ERROR; 
	} 
	area_info bm_info; 
	
	if (get_area_info(my_bitmap->Area(), &bm_info) != B_OK) { 
		return B_ERROR; 
	} 
	
	buffer_clone_info bc_info; 
	bc_info.area = bm_info.area; 
	bc_info.offset = ((char *) my_bitmap->Bits())-((char *) bm_info.address); 
	bc_info.size = my_bitmap->BitsLength(); 
	
	BBuffer *out_buffer = NULL; 
	status_t err = my_group->AddBuffer(bc_info, &out_buffer); 
*/	
	/* out_buffer is now set to use the BBitmap's memory */
	//return my_group;
	return NULL;
}

void GraphicsDevice::DeleteBufferGroup(BBufferGroup * uBufferGroup) {
}

image_id GraphicsDevice::LoadAccelerantImage() {
	status_t result;
	image_id image = -1;
	uint i;
	char
		signature[1024],
		path[PATH_MAX];
	struct stat st;
	static const directory_which vols[] = {
		B_USER_ADDONS_DIRECTORY,
		B_COMMON_ADDONS_DIRECTORY,
		B_BEOS_ADDONS_DIRECTORY
	};

	//get the accelerant signature from the driver...
	result = ioctl(hDevice, B_GET_ACCELERANT_SIGNATURE, &signature, sizeof(signature));
	if (result != B_OK) goto done;

	for (i=0; i < sizeof (vols) / sizeof (vols[0]); i++) {
		//	compute directory path to common or beos addon directory on
		//	floppy or boot volume

  	if (find_directory (vols[i], -1, false, path, PATH_MAX) != B_OK) continue;

		strcat (path, "/accelerants/");
		strcat (path, signature);

		// don't try to load non-existant files
		if (stat(path, &st) != 0) continue;

		// load the image
		image = load_add_on(path);
		if (image >= 0) {

			// get entrypoint from accelerant
			result = get_image_symbol (
				image,
				B_ACCELERANT_ENTRY_POINT,
#if defined(__INTEL__)
				B_SYMBOL_TYPE_ANY,
#else
				B_SYMBOL_TYPE_TEXT,
#endif
				(void **)&GetAccelerantFunction
			);

			if (result == B_OK) {
				init_accelerant ia;
				ia = (init_accelerant)GetAccelerantFunction(B_INIT_ACCELERANT, NULL);
				if (ia && ((result = ia(hDevice)) == B_OK)) {
					// we have a winner!
					break;
				}
			}

			// unload the accelerant, as we must be able to init!
			unload_add_on(image); 
		}

		// mark failure to load image
		image = -1;
	}

done:
	return image;
}


status_t GraphicsDevice::GetDisplayMode( display_mode * uDisplayMode ) {
	get_display_mode Accelerant_GetDisplayMode =
		(get_display_mode)GetAccelerantFunction(B_GET_DISPLAY_MODE, NULL);
		
	return Accelerant_GetDisplayMode( uDisplayMode );
}


status_t GraphicsDevice::SetDisplayMode( display_mode * new_mode, bool = false ) {
	propose_display_mode pdm;
	pdm = (propose_display_mode)GetAccelerantFunction(B_PROPOSE_DISPLAY_MODE, NULL);
	display_mode low, high;
	
	//KP: The following switch was added because I determined, through testing,
	//that propose_mode crashes on certain drivers if called without a valid color space.
	switch (new_mode->space) {
	case B_GRAY8:
		new_mode->space = B_CMAP8;
		break;
	//The following modes have been tested to work on my system...
	case B_CMAP8:
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
	case B_RGB24:
	case B_RGB32:
	case B_RGBA32:
		break;
	default:
		new_mode->space = B_CMAP8;
		break;
	}	
	
	low = *new_mode;
	new_mode->virtual_height = 0xFFFF;  //Make the virtual height as tall as possible
	high = *new_mode;
	if (pdm != NULL) {
		if (pdm( new_mode, &low, &high ) != B_ERROR) {
			set_display_mode sdm;
			sdm = (set_display_mode)GetAccelerantFunction(B_SET_DISPLAY_MODE, NULL);
			if (sdm != NULL) {
				return sdm(new_mode);
				printf("New Mode Set OK!");
			}
		}
	}
	return B_ERROR;
}
